#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = ['ClientHandlers', 'PageItems', 'nicegui_page_sync_coro', 'sync_like_page', 
           'sl_page', 'nicegui_page_async_coro', 'async_page', 'apage',
           'nicegui_page_class_async_coro', 'async_page_class', 'apage_class', 
           'nicegui_event_handler_func_async_coro', 'async_event_handler_func', 
           'aevent_handler_func', 'nicegui_event_handler_method_async_coro', 
           'async_event_handler_method', 'aevent_handler_method', 'PageContextBase', 
           'run']


import asyncio


class CurrentTask:
    def __init__(self, original) -> None:
        self.original = original
        self.new = original

    def patch(self, new):
        self.new = new

    def restore(self):
        self.new = self.original

    def __call__(self, *args, **kwds):
        return self.new(*args, **kwds)


current_task = CurrentTask(asyncio.current_task)
asyncio.current_task = current_task


from cengal.parallel_execution.coroutines.integrations.uvloop import uvloop_install
uvloop_install()


from cengal.parallel_execution.asyncio.init_loop import init_loop
init_loop()


if True:
    # increasing max decode packets to be able to transfer images
    # see https://github.com/miguelgrinberg/python-engineio/issues/142
    from engineio.payload import Payload
    Payload.max_decode_packets = 500


import inspect
from time import perf_counter
from typing import Set, Tuple, Optional, Union, Dict, Callable, Any, Hashable, cast, List, Coroutine
import warnings
from cengal.parallel_execution.asyncio.atasks import create_task
from cengal.parallel_execution.coroutines.coro_scheduler import Interface, AnyWorker, cs_acoro
from cengal.parallel_execution.coroutines.coro_standard_services.async_event_bus import (AsyncEventBus,
                                                                                         AsyncEventBusRequest)
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import PutCoro, put_coro_to
from cengal.parallel_execution.coroutines.coro_standard_services.run_coro import RunCoro, arun_coro_fast, run_coro_fast
from cengal.parallel_execution.coroutines.coro_standard_services.db import Db, DbRequest, EnvId, EnvInfo
from cengal.parallel_execution.coroutines.coro_standard_services.simple_yield import Yield
from cengal.parallel_execution.coroutines.coro_standard_services.sleep import Sleep
from cengal.parallel_execution.coroutines.coro_standard_services.asyncio_loop import AsyncioLoop, AsyncioLoopRequest
from cengal.parallel_execution.coroutines.coro_standard_services.shutdown_loop import ShutdownLoop
from cengal.parallel_execution.coroutines.coro_standard_services.shutdown_on_keyboard_interrupt import ShutdownOnKeyboardInterrupt
from cengal.parallel_execution.coroutines.coro_standard_services.simple_yield import Yield
from cengal.parallel_execution.coroutines.coro_standard_services.instance import Instance, InstanceRequest, afast_wait, fast_get_explicit, fast_set_explicit
from cengal.parallel_execution.coroutines.coro_standard_services.loop_yield import agly_patched
from cengal.parallel_execution.coroutines.coro_standard_services.log import LogRequest
from cengal.parallel_execution.coroutines.coro_tools.await_coro import await_coro_prim
from cengal.parallel_execution.coroutines.coro_tools.run_in_loop import (arun_in_fast_loop, arun_in_loop, run_in_loop)
from cengal.parallel_execution.coroutines.coro_tools.lock import Lock
from cengal.code_flow_control.args_manager import args_kwargs, EntityArgsHolder, EntityArgsHolderExplicit
from cengal.io.serve_free_ports import simple_port_search
from uuid import uuid4
from inspect import signature, Signature, isclass
from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.data_manipulation.conversion.reinterpret_cast import \
    reinterpret_cast
from cengal.parallel_execution.coroutines.coro_scheduler import (
    CoroWrapperAsyncAwait, Interface)
from cengal.parallel_execution.coroutines.coro_tools.await_coro import \
    await_coro_prim
from cengal.parallel_execution.coroutines.coro_tools.wait_coro import sync_coro_param
from cengal.parallel_execution.asyncio.atasks import create_task


from nicegui import app, ui, Client
# from nicegui.binding import loop
# agly_patched(loop)
from cengal.parallel_execution.asyncio.timed_yield import TimedYield
import nicegui.binding
original__propagate = nicegui.binding.propagate
class PropagateMock:
    def __init__(self) -> None:
        self.ty: TimedYield = TimedYield(0.01)
    
    def __call__(self, source_obj: Any, source_name: str, visited: Optional[Set[Tuple[int, str]]] = None) -> None:
        return original__propagate(source_obj, source_name, visited)


nicegui.binding.propagate = PropagateMock()
from nicegui.events import handle_event, GenericEventArguments, UiEventArguments
from nicegui.dataclasses import KWONLY_SLOTS
from nicegui.slot import Slot
from nicegui.globals import slot_stacks
from fastapi import Request as FastAPIRequest
from .text_translation import setup_translation, create_translatable_text_element, TextTranslator, \
    TranslationLanguageMapper, TranslationLanguageChooser, TranslatableTextElement, TTE, NiceguiTranslatableTextElement, NTTE
from cengal.parallel_execution.coroutines.coro_tools.await_coro import asyncio_coro
from cengal.file_system.app_fs_structure.app_dir_path import AppDirPath, AppDirectoryType
from cengal.file_system.path_manager import RelativePath
from cengal.web_tools.detect_browsers_host_device_type.by_http_headers import ClientViewType, client_view_type
from cengal.web_tools.detect_browsers_language.by_http_headers import parse_accept_language, optimize_accept_language, match_langs, normalize_lang
from cengal.math.numbers import RationalNumber
from cengal.base.exceptions import CengalError
from cengal.introspection.inspect import entity_name, entity_repr
from functools import partial
from collections import OrderedDict
from datetime import datetime, timedelta
from starlette.datastructures import Address, URL
from starlette.responses import Response, RedirectResponse
from aioipinfo import IPInfoClient, IPInfoError, IPInfoResponse
from dataclasses import dataclass
from contextlib import nullcontext
import os
import hashlib
import logging


DESTROY_CS_EVENT = f'DESTROY_CS_EVENT__{uuid4()}'
CS_PREPARED_FLAG = f'CS_PREPARED_FLAG'
CS_DESTROY_TIMEOUT = 3.0


text_translator: TextTranslator = None
translation_language_mapper: TranslationLanguageMapper = None
session_clients: Dict[Hashable, Set[Hashable]] = dict()
translatable_text_element_per_session: Dict[Hashable, NiceguiTranslatableTextElement] = dict()
translatable_text_element_per_client: Dict[Hashable, NiceguiTranslatableTextElement] = dict()

known_translation_param_names: Set[str] = {'_t', '_T', '__'}
known_page_context_param_names: Set[str] = {'page_context', 'pc', 'self'}
db_env_id: EnvId = 'cengal_nicegui_db_env'
db_request__logged_in_clients: DbRequest = DbRequest(db_env_id, 'logged_in_clients')
db_request__user_clients: DbRequest = DbRequest(db_env_id, 'user_clients')
db_request__logged_in_sessions: DbRequest = DbRequest(db_env_id, 'logged_in_sessions')
db_request__user_sessions: DbRequest = DbRequest(db_env_id, 'user_sessions')
db_request__ip_geolocation: DbRequest = DbRequest(db_env_id, 'ip_geolocation')
db_request__user_sign_in_info: DbRequest = DbRequest(db_env_id, 'user_sign_in_info')
db_request__credentials_by_user_id: DbRequest = DbRequest(db_env_id, 'credentials_by_user_id')
db_request__user_sign_up_info: DbRequest = DbRequest(db_env_id, 'user_sign_up_info')
signed_in_session_timeout: timedelta = timedelta(days=90)
ipinfo_io_token: str = os.environ.get('CENGAL_NICEGUI__IPINFO_IO_TOKEN', None)


class CengalNiceguiException(CengalError):
    pass


class UserNotFoundError(CengalNiceguiException):
    pass


class PasswordIsNotCorrectError(CengalNiceguiException):
    pass


class CoroWrapperNiceGuiPageAsyncAwait(CoroWrapperAsyncAwait):
    def reinit_for_nicegui(self):
        self.subtype = self._setup_subtype()
        self._current_call_method = self._call_method  # type: Callable
        self.original__current_task = None

    def mock__current_task(self, loop=None):
        return self.current_asyncio_task

    def patch_asyncio_current_task(self):
        current_task.patch(self.mock__current_task)

    def unpatch_asyncio_current_task(self):
        current_task.restore()

    def _init_coroutine(self, *init_args, **init_kwargs):
        try:
            self.patch_asyncio_current_task()
            return super()._init_coroutine(*init_args, **init_kwargs)
        finally:
            self.unpatch_asyncio_current_task()

    def _call_coroutine(self, *args, **kwargs):
        try:
            self.patch_asyncio_current_task()
            return super()._call_coroutine(*args, **kwargs)
        finally:
            self.unpatch_asyncio_current_task()

    def _throw_coroutine(self, ex_type, ex_value=None, ex_traceback=None):
        try:
            self.patch_asyncio_current_task()
            return super()._throw_coroutine(ex_type, ex_value, ex_traceback)
        finally:
            self.unpatch_asyncio_current_task()

    def _close_coroutine(self, *args, **kwargs):
        try:
            self.patch_asyncio_current_task()
            return super()._close_coroutine(*args, **kwargs)
        finally:
            self.unpatch_asyncio_current_task()


class CoroWrapperNiceGuiPageGreenlet(CoroWrapperGreenlet):
    def reinit_for_nicegui(self):
        self.subtype = self._setup_subtype()
        self._current_call_method = self._call_method  # type: Callable
        self.original__current_task = None

    def mock__current_task(self, loop=None):
        return self.current_asyncio_task

    def patch_asyncio_current_task(self):
        current_task.patch(self.mock__current_task)

    def unpatch_asyncio_current_task(self):
        current_task.restore()

    def _init_coroutine(self, *init_args, **init_kwargs):
        try:
            self.patch_asyncio_current_task()
            return super()._init_coroutine(*init_args, **init_kwargs)
        finally:
            self.unpatch_asyncio_current_task()

    def _call_coroutine(self, *args, **kwargs):
        try:
            self.patch_asyncio_current_task()
            return super()._call_coroutine(*args, **kwargs)
        finally:
            self.unpatch_asyncio_current_task()

    def _throw_coroutine(self, ex_type, ex_value=None, ex_traceback=None):
        try:
            self.patch_asyncio_current_task()
            return super()._throw_coroutine(ex_type, ex_value, ex_traceback)
        finally:
            self.unpatch_asyncio_current_task()

    def _close_coroutine(self, *args, **kwargs):
        try:
            self.patch_asyncio_current_task()
            return super()._close_coroutine(*args, **kwargs)
        finally:
            self.unpatch_asyncio_current_task()


class ClientHandlers:
    def __init__(self, client: Optional[Client]) -> None:
        self.client: Optional[Client] = client
        self.on_connected_handlers: Set[Callable] = set()
        self.on_disconnected_handlers: Set[Callable] = set()
        self.is_connected: bool = False
        self.is_disconnected: bool = False
    
    @classmethod
    def check_args_factory(cls, signature: Signature, args, kwargs) -> Optional['ClientHandlers']:
        client: Optional[Client] = kwargs.get('client', None)
        if (client is None) or (not isinstance(client, Client)):
            return None
        else:
            return cls.install(client)
    
    @classmethod
    def install(cls, client: Client) -> 'ClientHandlers':
        client_handlers: ClientHandlers = cls(client)
        client.handlers = client_handlers
        client.on_connect(client_handlers._on_connected())
        client.on_disconnect(client_handlers._on_disconnected())
        return client_handlers
    
    def add_on_connected_handler(self, handler: Callable):
        if self.is_connected:
            handler()
        else:
            self.on_connected_handlers.add(handler)
    
    def remove_on_connected_handler(self, handler: Callable):
        self.on_connected_handlers.discard(handler)
    
    async def _on_connected(self):
        on_connected_handlers_buff = tuple(self.on_connected_handlers)
        self.on_connected_handlers = set()
        for index, handler in enumerate(on_connected_handlers_buff):
            try:
                handler()
            except:
                self.on_connected_handlers.update(on_connected_handlers_buff[index + 1:])
                create_task(self._on_connected)
                raise
    
    def add_on_disconnected_handler(self, handler: Callable):
        if self.is_disconnected:
            handler()
        else:
            self.on_disconnected_handlers.add(handler)
    
    def remove_on_disconnected_handler(self, handler: Callable):
        self.on_disconnected_handlers.discard(handler)
    
    async def _on_disconnected(self):
        on_disconnected_handlers_buff = tuple(self.on_disconnected_handlers)
        self.on_disconnected_handlers = set()
        for index, handler in enumerate(on_disconnected_handlers_buff):
            try:
                handler()
            except:
                self.on_disconnected_handlers.update(on_disconnected_handlers_buff[index + 1:])
                create_task(self._on_disconnected)
                raise


class PageItems:
    def __init__(self, client: Optional[Client]) -> None:
        self.client: Optional[Client] = client
        self.items: Set[Any] = set()
    
    @classmethod
    def check_args_factory(cls, signature: Signature, args, kwargs) -> Optional['PageItems']:
        client: Optional[Client] = kwargs.get('client', None)
        if (client is None) or (not isinstance(client, Client)):
            return None
        else:
            return cls.install(client)
    
    @classmethod
    def install(cls, client: Client) -> 'PageItems':
        page_items: PageItems = cls(client)
        client.page_items = page_items
        return page_items
    
    def add(self, item: Any):
        self.items.add(item)
    
    def remove(self, item: Any):
        self.items.discard(item)
    
    def __call__(self, item: Any) -> Any:
        return self.add(item)


async def get_ip_geolocation(ipinfo_io_token, host):
    async with IPInfoClient(ipinfo_io_token) as client:
        try:
            ip_geolocation = await client.ipinfo_dict(host)
            if ip_geolocation is not None:
                await i(db_request__ip_geolocation.put(host))
        except IPInfoError:
            ip_geolocation = None
        
        return ip_geolocation


def nicegui_event_handler_func_async_coro(coro_worker: Worker):
    """Decorator. Without arguments. Makes a proper, fully functional async Page from any decorated Cengal coroutine

    Args:
        coro_worker (Worker): _description_

    Raises:
        TypeError: _description_

    Returns:
        _type_: _description_
    """    
    coro_worker_0 = coro_worker
    # async def wrapper(*args, **kwargs):
    async def wrapper(*args, **kwargs):
        await await_cs_initiated()
        coro_worker = coro_worker_0
        app.cs.logger.debug(f'nicegui_event_handler_async_coro.wrapper - start: {entity_name(coro_worker)}')
        coro_worker_type: CoroType = find_coro_type(coro_worker)
        if CoroType.awaitable == coro_worker_type:
            async def awaitable_coro_wrapper(i: Interface, current_asyncio_task, 
                                             coro_worker_with_args: EntityArgsHolderExplicit):
                coro_worker, args, kwargs = coro_worker_with_args.entity_args_kwargs()
                i.log.debug(f'nicegui_event_handler_async_coro.wrapper.awaitable_coro_wrapper - start: {entity_name(coro_worker)}')
                await apretent_to_be_asyncio_coro(i, current_asyncio_task)
                # await i(Instance, InstanceRequest().wait(CS_PREPARED_FLAG))
                i.log.debug(f'nicegui_event_handler_async_coro.wrapper.awaitable_coro_wrapper - ready: {entity_name(coro_worker)}')
                return await coro_worker(i, *args, **kwargs)
            
            coro_wrapper = awaitable_coro_wrapper
        elif CoroType.greenlet == coro_worker_type:
            def greenlet_coro_wrapper(i: Interface, current_asyncio_task, 
                                      coro_worker_with_args: EntityArgsHolderExplicit):
                coro_worker, args, kwargs = coro_worker_with_args.entity_args_kwargs()
                i.log.debug(f'nicegui_event_handler_async_coro.wrapper.greenlet_coro_wrapper - start: {entity_name(coro_worker)}')
                pretent_to_be_asyncio_coro(i, current_asyncio_task)
                # i(Instance, InstanceRequest().wait(CS_PREPARED_FLAG))
                i.log.debug(f'nicegui_event_handler_async_coro.wrapper.greenlet_coro_wrapper - ready: {entity_name(coro_worker)}')
                return coro_worker(i, *args, **kwargs)
            
            coro_wrapper = greenlet_coro_wrapper
        else:
            raise TypeError(f'{coro_worker} is neither an awaitable nor a greenlet')

        current_asyncio_task = asyncio.current_task()
        return await await_coro_prim(coro_wrapper, current_asyncio_task, 
                                     EntityArgsHolderExplicit(coro_worker, args, kwargs))
        
    coro_worker_sign: Signature = signature(coro_worker)
    wrapper_signature = coro_worker_sign.replace(parameters=tuple(coro_worker_sign.parameters.values())[1:], return_annotation=coro_worker_sign.return_annotation)
    wrapper.__signature__ = wrapper_signature
    
    return wrapper


async_event_handler_func = nicegui_event_handler_func_async_coro
aevent_handler_func = nicegui_event_handler_func_async_coro


def nicegui_event_handler_method_async_coro(coro_worker: Worker):
    """Decorator. Without arguments. Makes a proper, fully functional async Page from any decorated Cengal coroutine

    Args:
        coro_worker (Worker): _description_

    Raises:
        TypeError: _description_

    Returns:
        _type_: _description_
    """    
    coro_worker_0 = coro_worker
    # async def wrapper(*args, **kwargs):
    async def wrapper(self, *args, **kwargs):
        await await_cs_initiated()
        coro_worker = coro_worker_0
        coro_worker_name: str = coro_worker.__name__
        unwrapped_coro_worker_name: str = f'_unwrapped_method__{coro_worker_name}'
        if not hasattr(self, unwrapped_coro_worker_name):
            bound_coro_worker = coro_worker.__get__(self, self.__class__)
            setattr(self, unwrapped_coro_worker_name, bound_coro_worker)

        coro_worker = getattr(self, unwrapped_coro_worker_name)
        app.cs.logger.debug(f'nicegui_event_handler_async_coro.wrapper - start: {entity_name(coro_worker)}')
        coro_worker_type: CoroType = find_coro_type(coro_worker)
        if CoroType.awaitable == coro_worker_type:
            async def awaitable_coro_wrapper(i: Interface, current_asyncio_task, 
                                             coro_worker_with_args: EntityArgsHolderExplicit):
                coro_worker, args, kwargs = coro_worker_with_args.entity_args_kwargs()
                i.log.debug(f'nicegui_event_handler_async_coro.wrapper.awaitable_coro_wrapper - start: {entity_name(coro_worker)}')
                await apretent_to_be_asyncio_coro(i, current_asyncio_task)
                # await i(Instance, InstanceRequest().wait(CS_PREPARED_FLAG))
                i.log.debug(f'nicegui_event_handler_async_coro.wrapper.awaitable_coro_wrapper - ready: {entity_name(coro_worker)}')
                i.log.debug(f'{entity_repr(coro_worker)}')
                return await coro_worker(i, *args, **kwargs)
            
            coro_wrapper = awaitable_coro_wrapper
        elif CoroType.greenlet == coro_worker_type:
            def greenlet_coro_wrapper(i: Interface, current_asyncio_task, 
                                      coro_worker_with_args: EntityArgsHolderExplicit):
                coro_worker, args, kwargs = coro_worker_with_args.entity_args_kwargs()
                i.log.debug(f'nicegui_event_handler_async_coro.wrapper.greenlet_coro_wrapper - start: {entity_name(coro_worker)}')
                pretent_to_be_asyncio_coro(i, current_asyncio_task)
                # i(Instance, InstanceRequest().wait(CS_PREPARED_FLAG))
                i.log.debug(f'nicegui_event_handler_async_coro.wrapper.greenlet_coro_wrapper - ready: {entity_name(coro_worker)}')
                return coro_worker(i, *args, **kwargs)
            
            coro_wrapper = greenlet_coro_wrapper
        else:
            raise TypeError(f'{coro_worker} is neither an awaitable nor a greenlet')

        current_asyncio_task = asyncio.current_task()
        return await await_coro_prim(coro_wrapper, current_asyncio_task, 
                                     EntityArgsHolderExplicit(coro_worker, args, kwargs))
        
    coro_worker_sign: Signature = signature(coro_worker)
    new_parameters = list(coro_worker_sign.parameters.values())
    del new_parameters[1]
    wrapper_signature = coro_worker_sign.replace(parameters=tuple(new_parameters), return_annotation=coro_worker_sign.return_annotation)
    wrapper.__signature__ = wrapper_signature
    
    return wrapper


async_event_handler_method = nicegui_event_handler_method_async_coro
aevent_handler_method = nicegui_event_handler_method_async_coro


def get_task_slot_stack(task_id) -> List[Slot]:
    if task_id not in slot_stacks:
        slot_stacks[task_id] = []
    return slot_stacks[task_id]


class FakeElementForEvents:
    def __init__(self, task_id) -> None:
        self.is_ignoring_events: bool = False
        self.parent_slot: Union[Slot, nullcontext] = nullcontext()
        slot_stack = get_task_slot_stack(task_id)
        if slot_stack:
            self.parent_slot = slot_stack[-1]


@dataclass(**KWONLY_SLOTS)
class OnSignInEventArguments(UiEventArguments):
    user_id: Hashable
    user_sign_in_info: Union[Dict, Tuple, List]


@dataclass(**KWONLY_SLOTS)
class OnSignOutEventArguments(UiEventArguments):
    user_id: Hashable


class PageContextBase:
    def __init__(self, client: Client, request: FastAPIRequest, _t: NTTE, current_asyncio_task) -> None:
        self.client: Client = client
        self.request: FastAPIRequest = request
        self.client_host_port: Address = self.request.client
        self.client_id: Hashable = client.id
        self.session_id: Hashable = client.session_id
        self.user_id: Hashable = None
        self._t: NTTE = _t
        self.current_asyncio_task = current_asyncio_task
        self.current_asyncio_task_id = id(current_asyncio_task)
        self.client_view_type: ClientViewType = self._determine_client_view_type()
        self.better_lang: str = None
        self.featured_langs: OrderedDict[str, RationalNumber] = None
        self.langs: OrderedDict[str, RationalNumber] = None
        self.better_lang, self.featured_langs, self.langs = self._determine_client_languages()
        self._t.text_translation_language_chooser.lang = self.better_lang
    
    def _init(self, i: Interface):
        self.register_session_page(i)
        i(PutCoro, self._arequest_ip_geolocation, self.client_host_port[0])
        if self._find_user(i):
            i.log.debug(f'{self.session_id}, {self.client_id}: logged in')
        else:
            i.log.debug(f'{self.session_id}, {self.client_id}: logged out')
    
    async def _ainit(self, i: Interface):
        self.register_session_page(i)
        await i(PutCoro, self._arequest_ip_geolocation, self.client_host_port[0])
        if await self._afind_user(i):
            i.log.debug(f'{self.session_id}, {self.client_id}: logged in')
        else:
            i.log.debug(f'{self.session_id}, {self.client_id}: logged out')
    
    async def _arequest_ip_geolocation(self, i: Interface, host: str):
        ip_geolocation: Optional[Dict] = None
        need_to_request: bool = True
        try:
            ip_geolocation = await i(db_request__ip_geolocation.get(host))
            need_to_request = False
        except KeyError:
            pass

        if need_to_request:
            ip_geolocation = await i(AsyncioLoop, AsyncioLoopRequest().wait(get_ip_geolocation(ipinfo_io_token, host)))
        
        await self.on_ip_geolocation_ready(i, ip_geolocation)
    
    def register_session_page(self, i: Interface):
        session_pages: Dict[Hashable, Set[PageContextBase]] = fast_get_explicit(i, 'cengal_nicegui__session_pages')
        if self.session_id not in session_pages:
            session_pages[self.session_id] = set()
        
        session_pages[self.session_id].add(self)
    
    def _determine_client_view_type(self) -> ClientViewType:
        return client_view_type(self.request.headers)
    
    def _determine_client_languages(self):
        parsed_accept_language: Optional[OrderedDict[str, RationalNumber]] = optimize_accept_language(parse_accept_language(self.request.headers))
        translation_data: Dict = self._t.text_translator.decoded_data
        return match_langs(
                translation_data['default_language'], 
                set(translation_data['featured_languages']),
                set(translation_data['supported_languages']),
                translation_data['translation_language_map'],
                parsed_accept_language,
            )
    
    def _find_user(self, i: Interface) -> bool:
        if self.session_id is None:
            return False
        
        try:
            user_id, sign_in_date_time = i(db_request__logged_in_sessions.get(self.session_id))
            sign_in_date_time = datetime.fromisoformat(sign_in_date_time)
            if (datetime.now() - sign_in_date_time) > signed_in_session_timeout:
                i(db_request__logged_in_sessions.delete(self.session_id))
                return False
            
            self.user_id = user_id
            return True
        except KeyError:
            return False
    
    async def _afind_user(self, i: Interface) -> bool:
        if self.session_id is None:
            return False
        
        try:
            user_id, sign_in_date_time = await i(db_request__logged_in_sessions.get(self.session_id))
            sign_in_date_time = datetime.fromisoformat(sign_in_date_time)
            if (datetime.now() - sign_in_date_time) > signed_in_session_timeout:
                await i(db_request__logged_in_sessions.delete(self.session_id))
                return False
            
            self.user_id = user_id
            return True
        except KeyError:
            return False
    
    # def find_user(self, i: Interface, client_connection_check_interval: RationalNumber = 0.01) -> bool:
    #     if not self.client.has_socket_connection:
    #         if not self.client.is_waiting_for_connection:
    #             i(AsyncioLoop, AsyncioLoopRequest().wait(self.client.connected()))
        
    #     while self.client.is_waiting_for_connection:
    #         i(Sleep, client_connection_check_interval)
        
    #     return self._find_user(i)
    
    # async def afind_user(self, i: Interface, client_connection_check_interval: RationalNumber = 0.01) -> bool:
    #     if not self.client.has_socket_connection:
    #         if not self.client.is_waiting_for_connection:
    #             await i(AsyncioLoop, AsyncioLoopRequest().wait(self.client.connected()))
        
    #     while self.client.is_waiting_for_connection:
    #         await i(Sleep, client_connection_check_interval)
        
    #     return await self._afind_user(i)
    
    def try_sign_in(self, i: Interface, credentials: Union[Dict, Tuple, List, str, bytes, int, float]) -> Optional[Union[Dict, Tuple, List]]:
        user_data: Union[Dict, Tuple, List] = None

        try:
            user_data = i(db_request__user_sign_in_info.get(credentials))
        except KeyError:
            pass
        
        if user_data is None:
            return None
        
        user_id = user_data[0]
        if self.session_id is None:
            i(db_request__logged_in_clients.put(self.client_id, (user_id, datetime.now())))
            user_clients_lock: Lock = fast_get_explicit(i, 'cengal_nicegui__user_clients_lock')
            with user_clients_lock:
                try:
                    user_clients: Set[Hashable] = set(i(db_request__user_clients.get(user_id)))
                except KeyError:
                    user_clients = set()
                
                user_clients.add(self.client_id)
                i(db_request__user_sessions.put(user_id, user_clients))
            
            i(PutCoro, self._on_signed_in, user_id)
        else:
            i(db_request__logged_in_sessions.put(self.session_id, (user_id, datetime.now())))
            user_sessions_lock: Lock = fast_get_explicit(i, 'cengal_nicegui__user_sessions_lock')
            with user_sessions_lock:
                try:
                    user_sessions: Set[Hashable] = set(i(db_request__user_sessions.get(user_id)))
                except KeyError:
                    user_sessions = set()
                
                user_sessions.add(self.session_id)
                i(db_request__user_sessions.put(user_id, user_sessions))
            
            session_pages: Dict[Hashable, Set[PageContextBase]] = fast_get_explicit(i, 'cengal_nicegui__session_pages')
            current_session_pages: Set[PageContextBase] = session_pages.get(self.session_id, set())
            for session_page in current_session_pages:
                i(PutCoro, session_page._on_signed_in, user_id, user_data)
        
        return user_data
    
    async def atry_sign_in(self, i: Interface, credentials: Union[Dict, Tuple, List, str, bytes, int, float]) -> Optional[Union[Dict, Tuple, List]]:
        user_data: Union[Dict, Tuple, List] = None

        try:
            user_data = await i(db_request__user_sign_in_info.get(credentials))
        except KeyError:
            pass
        
        if user_data is None:
            return None
        
        user_id = user_data[0]
        if self.session_id is None:
            await i(db_request__logged_in_clients.put(self.client_id, (user_id, datetime.now())))
            user_clients_lock: Lock = fast_get_explicit(i, 'cengal_nicegui__user_clients_lock')
            async with user_clients_lock:
                try:
                    user_clients: Set[Hashable] = set(await i(db_request__user_clients.get(user_id)))
                except KeyError:
                    user_clients = set()
                
                user_clients.add(self.client_id)
                await i(db_request__user_sessions.put(user_id, user_clients))
            
            await i(PutCoro, self._on_signed_in, user_id)
        else:
            await i(db_request__logged_in_sessions.put(self.session_id, (user_id, datetime.now())))
            user_sessions_lock: Lock = fast_get_explicit(i, 'cengal_nicegui__user_sessions_lock')
            async with user_sessions_lock:
                try:
                    user_sessions: Set[Hashable] = set(await i(db_request__user_sessions.get(user_id)))
                except KeyError:
                    user_sessions = set()
                
                user_sessions.add(self.session_id)
                await i(db_request__user_sessions.put(user_id, user_sessions))
            
            session_pages: Dict[Hashable, Set[PageContextBase]] = fast_get_explicit(i, 'cengal_nicegui__session_pages')
            current_session_pages: Set[PageContextBase] = set(session_pages.get(self.session_id, set()))  # TODO: robust fix needed for `RuntimeError: Set changed size during iteration`
            for session_page in current_session_pages:
                await i(PutCoro, session_page._on_signed_in, user_id, user_data)
        
        return user_data
    
    def user_signed_up(self, i: Interface, credentials: Union[Dict, Tuple, List, str, bytes, int, float],
                              user_sign_in_info: Union[Dict, Tuple, List]):
        i(db_request__user_sign_in_info.put(credentials, user_sign_in_info))
    
    async def auser_signed_up(self, i: Interface, credentials: Union[Dict, Tuple, List, str, bytes, int, float],
                              user_sign_in_info: Union[Dict, Tuple, List]):
        await i(db_request__user_sign_in_info.put(credentials, user_sign_in_info))

    def sign_out(self, i: Interface) -> bool:
        user_id = self.user_id
        if user_id is None:
            return False
        
        if self.session_id is None:
            try:
                i(db_request__logged_in_clients.delete(self.client_id))
            except KeyError:
                pass

            i(PutCoro, self._on_signed_out)
            return True
        else:
            try:
                i(db_request__logged_in_sessions.delete(self.session_id))
            except KeyError:
                pass

            session_pages: Dict[Hashable, Set[PageContextBase]] = fast_get_explicit(i, 'cengal_nicegui__session_pages')
            current_session_pages: Set[PageContextBase] = session_pages.get(self.session_id, set())
            for session_page in current_session_pages:
                i(PutCoro, session_page._on_signed_out)

            return True

    async def asign_out(self, i: Interface) -> bool:
        user_id = self.user_id
        if user_id is None:
            return False
        
        if self.session_id is None:
            try:
                await i(db_request__logged_in_clients.delete(self.client_id))
            except KeyError:
                pass

            user_clients_lock: Lock = fast_get_explicit(i, 'cengal_nicegui__user_clients_lock')
            async with user_clients_lock:
                try:
                    user_clients: Set[Hashable] = set(await i(db_request__user_clients.get(user_id)))
                except KeyError:
                    user_clients = set()
                
                user_clients.discard(self.client_id)
                await i(db_request__user_sessions.put(user_id, user_clients))

            await i(PutCoro, self._on_signed_out)
            return True
        else:
            try:
                await i(db_request__logged_in_sessions.delete(self.session_id))
            except KeyError:
                pass

            user_sessions_lock: Lock = fast_get_explicit(i, 'cengal_nicegui__user_sessions_lock')
            async with user_sessions_lock:
                try:
                    user_sessions: Set[Hashable] = set(await i(db_request__user_sessions.get(user_id)))
                except KeyError:
                    user_sessions = set()
                
                user_sessions.discard(self.session_id)
                await i(db_request__user_sessions.put(user_id, user_sessions))

            session_pages: Dict[Hashable, Set[PageContextBase]] = fast_get_explicit(i, 'cengal_nicegui__session_pages')
            current_session_pages: Set[PageContextBase] = session_pages.get(self.session_id, set())
            for session_page in current_session_pages:
                await i(PutCoro, session_page._on_signed_out)

            return True

    def try_sign_up__login_password(self, i: Interface, login: str, password: str, 
                                    user_sign_in_info_maker: Callable[[Interface, str, Hashable, datetime], Union[Dict, Tuple, List]]) -> bool:
        login_exists: bool = False
        try:
            i(db_request__user_sign_up_info.get(login))
            login_exists = True
        except KeyError:
            pass

        if login_exists:
            return False
        
        password_bytes = password.encode('utf-8')
        salt = os.urandom(16)
        key_length = 64
        N = 16384  # CPU/memory cost factor
        r = 8     # Block size
        p = 1     # Parallelization factor
        dk = hashlib.scrypt(password_bytes, salt=salt, n=N, r=r, p=p, dklen=key_length)
        i(db_request__user_sign_up_info.put(login, (dk, salt, N, r, p, key_length)))
        user_id: Hashable = uuid4()
        credentials: Tuple = (login, dk)
        i(db_request__credentials_by_user_id.put(user_id, credentials))
        user_sign_in_info: Union[Dict, Tuple, List] = run_coro_fast(i, user_sign_in_info_maker, login, user_id, datetime.now())
        self.user_signed_up(i, credentials, user_sign_in_info)
        return True

    async def atry_sign_up__login_password(self, i: Interface, login: str, password: str, 
                                           user_sign_in_info_maker: Callable[[Interface, str, Hashable, datetime], Union[Dict, Tuple, List]]) -> bool:
        login_exists: bool = False
        try:
            await i(db_request__user_sign_up_info.get(login))
            login_exists = True
        except KeyError:
            pass

        if login_exists:
            return False
        
        password_bytes = password.encode('utf-8')
        salt = os.urandom(16)
        key_length = 64
        N = 16384  # CPU/memory cost factor
        r = 8     # Block size
        p = 1     # Parallelization factor
        dk = hashlib.scrypt(password_bytes, salt=salt, n=N, r=r, p=p, dklen=key_length)
        await i(db_request__user_sign_up_info.put(login, (dk, salt, N, r, p, key_length)))
        user_id: Hashable = uuid4().bytes
        credentials: Tuple = (login, dk)
        await i(db_request__credentials_by_user_id.put(user_id, credentials))
        user_sign_in_info: Union[Dict, Tuple, List] = await arun_coro_fast(i, user_sign_in_info_maker, login, user_id, datetime.now())
        await self.auser_signed_up(i, credentials, user_sign_in_info)
        return True

    def try_sign_in__login_password(self, i: Interface, login: str, password: str) -> Union[Dict, Tuple, List]:
        user_found: bool = False
        try:
            dk, salt, N, r, p, key_length = i(db_request__user_sign_up_info.get(login))
            user_found = True
        except KeyError:
            pass

        if not user_found:
            raise UserNotFoundError

        password_bytes = password.encode('utf-8')
        dk_0 = hashlib.scrypt(password_bytes, salt=salt, n=N, r=r, p=p, dklen=key_length)
        if dk != dk_0:
            raise PasswordIsNotCorrectError
        
        credentials: Tuple = (login, dk)
        return self.try_sign_in(i, credentials)

    async def atry_sign_in__login_password(self, i: Interface, login: str, password: str) -> Union[Dict, Tuple, List]:
        user_found: bool = False
        try:
            dk, salt, N, r, p, key_length = await i(db_request__user_sign_up_info.get(login))
            user_found = True
        except KeyError:
            pass

        if not user_found:
            raise UserNotFoundError

        password_bytes = password.encode('utf-8')
        dk_0 = hashlib.scrypt(password_bytes, salt=salt, n=N, r=r, p=p, dklen=key_length)
        if dk != dk_0:
            raise PasswordIsNotCorrectError
        
        credentials: Tuple = (login, dk)
        return await self.atry_sign_in(i, credentials)

    def try_sign_up__email_password(self, i: Interface, email: str, password: str) -> bool:
        raise NotImplementedError

    async def atry_sign_up__email_password(self, i: Interface, email: str, password: str) -> bool:
        raise NotImplementedError

    async def _on_signed_in(self, i: Interface, user_id: Hashable, user_sign_in_info: Union[Dict, Tuple, List]) -> None:
        self.user_id = user_id
        handle_event(self._on_signed_in_impl, OnSignInEventArguments(sender=FakeElementForEvents(self.current_asyncio_task_id), client=self.client, user_id=user_id, user_sign_in_info=user_sign_in_info))

    @aevent_handler_method
    async def _on_signed_in_impl(self, i: Interface, arguments: OnSignInEventArguments) -> None:
        return await self.on_signed_in(i, arguments.user_id, arguments.user_sign_in_info)

    async def on_signed_in(self, i: Interface, user_id: Hashable, user_sign_in_info: Union[Dict, Tuple, List]) -> None:
        pass

    async def _on_signed_out(self, i: Interface) -> None:
        user_id = self.user_id
        self.user_id = None
        handle_event(self._on_signed_out_impl, OnSignOutEventArguments(sender=FakeElementForEvents(self.current_asyncio_task_id), client=self.client, user_id=user_id))

    @aevent_handler_method
    async def _on_signed_out_impl(self, i: Interface, arguments: OnSignOutEventArguments) -> None:
        return await self.on_signed_out(i, arguments.user_id)

    async def on_signed_out(self, i: Interface, user_id: Hashable) -> None:
        pass

    async def on_ip_geolocation_ready(self, i: Interface, ip_geolocation: Optional[Dict]) -> None:
        pass

    def refferer(self) -> Optional[str]:
        return self.request.headers.get('referer', None)

    def url(self) -> URL:
        return self.request.url

    def open_page(self, target: Optional[Union[Callable[..., Any], str]] = None, new_tab: bool = False) -> Optional[RedirectResponse]:
        if target is None:
            return
        
        client = self.client
        if client.has_socket_connection:
            client.open(target, new_tab)
        else:
            return RedirectResponse(target)

    def open_previous_page(self, default_target: Optional[Union[Callable[..., Any], str]] = None, new_tab: bool = False, can_be_current_page: bool = False) -> Optional[RedirectResponse]:
        target: str = self.refferer()
        current_url: str = str(self.url())
        if (target is None) or (False if can_be_current_page else (target == current_url)):
            if not isinstance(default_target, str):
                default_target = globals.page_routes[target]
            
            target = default_target if default_target is not None else '/'
        
        client = self.client
        if client.has_socket_connection:
            client.open(target, new_tab)
        else:
            return self.open_previous_page_http_response(default_target, can_be_current_page)

    def open_previous_page_http_response(self, default_target: Optional[Union[Callable[..., Any], str]] = None, can_be_current_page: bool = False) -> RedirectResponse:
        refferer: str = self.refferer()
        current_url: str = str(self.url())
        if (refferer is None) or (False if can_be_current_page else (refferer == current_url)):
            refferer = default_target if default_target is not None else '/'
        
        return RedirectResponse(refferer)


def nicegui_page_sync_coro(*dargs, **dkwargs):
    """Decorator. With an arguments. Gives ability to execute any decorated Cengal coroutine based page as a sync function.
    Can postpone execution to the actual loop when possible if None as an immediate result (no result) is acceptible.
    Can start own loop if needed. See sync_coro_param() decorator from cengal/parallel_execution/coroutines/coro_tools/wait_coro for more details

    Returns:
        _type_: _description_
    """    
    dargs_0 = dargs
    dkwargs_0 = dkwargs
    def nicegui_page_sync_coro_impl(coro_worker: Worker):
        coro_worker_0 = coro_worker
        dargs_1 = dargs_0
        dkwargs_1 = dkwargs_0
        def wrapper(*args, **kwargs):
            coro_worker = coro_worker_0
            dargs = dargs_1
            dkwargs = dkwargs_1
            sync_coro_decorator = sync_coro_param(*dargs, **dkwargs)
            sync_coro_decorator_wrapper = sync_coro_decorator(coro_worker)
            ClientHandlers.check_args_factory(wrapper_signature, args, kwargs)
            PageItems.check_args_factory(wrapper_signature, args, kwargs)
            return sync_coro_decorator_wrapper(*args, **kwargs)
            
        coro_worker_sign: Signature = signature(coro_worker)
        wrapper_signature = coro_worker_sign.replace(parameters=tuple(coro_worker_sign.parameters.values())[1:], return_annotation=coro_worker_sign.return_annotation)
        wrapper.__signature__ = wrapper_signature
        
        return wrapper
    return nicegui_page_sync_coro_impl


sync_like_page = nicegui_page_sync_coro
sl_page = nicegui_page_sync_coro


async def await_cs_initiated():
    while (not hasattr(app, 'cs_initiated')) or (not app.cs_initiated):
        await asyncio.sleep(0.01)


async def apretent_to_be_asyncio_coro(i: Interface, asyncio_coro):
    reinterpret_cast(CoroWrapperNiceGuiPageAsyncAwait, i._coro)
    i._coro.current_asyncio_task = asyncio_coro
    i._coro.reinit_for_nicegui()
    await i(Yield)


def pretent_to_be_asyncio_coro(i: Interface, asyncio_coro):
    reinterpret_cast(CoroWrapperNiceGuiPageGreenlet, i._coro)
    i._coro.current_asyncio_task = asyncio_coro
    i._coro.reinit_for_nicegui()
    i(Yield)


def nicegui_page_async_coro_impl(page_class: Optional[Union[PageContextBase, EntityArgsHolder]], coro_worker: Worker):
    """Decorator. Without arguments. Makes a proper, fully functional async Page from any decorated Cengal coroutine

    Args:
        coro_worker (Worker): _description_

    Raises:
        TypeError: _description_

    Returns:
        _type_: _description_
    """    
    coro_worker_0 = coro_worker
    page_class_0 = page_class
    # async def wrapper(*args, **kwargs):
    async def wrapper(request: FastAPIRequest, client: Client):
        await await_cs_initiated()
        request.app.cs.logger.debug(f'nicegui_page_async_coro_impl.wrapper - start: {request.url}')
        client_id = client.id
        session_id = request.session.get('id', None)
        client.session_id = session_id
        if session_id not in session_clients:
            session_clients[session_id] = set()
        
        session_clients[session_id].add(client_id)

        if session_id is None:
            translatable_text_element: NiceguiTranslatableTextElement = create_translatable_text_element(text_translator, translation_language_mapper)
            # translatable_text_element: NiceguiTranslatableTextElement = create_translatable_text_element(
            #         await asyncio_coro(cs_acoro(afast_wait))('text_translator'), await asyncio_coro(cs_acoro(afast_wait))('translation_language_mapper')
            #     )
            translatable_text_element_per_client[client_id] = translatable_text_element
        else:
            if session_id in translatable_text_element_per_session:
                translatable_text_element = translatable_text_element_per_session[session_id]
            else:
                translatable_text_element: NiceguiTranslatableTextElement = create_translatable_text_element(text_translator, translation_language_mapper)
                # translatable_text_element: NiceguiTranslatableTextElement = create_translatable_text_element(
                #         await asyncio_coro(cs_acoro(afast_wait))('text_translator'), await asyncio_coro(cs_acoro(afast_wait))('translation_language_mapper')
                #     )
                translatable_text_element_per_session[session_id] = translatable_text_element
        
        client.translatable_text_element = translatable_text_element
        await asyncio_coro(cs_acoro(translatable_text_element.aregister_on_lang_changed_handler))()
        coro_worker = coro_worker_0
        coro_worker_type: CoroType = find_coro_type(coro_worker)
        if CoroType.awaitable == coro_worker_type:
            async def awaitable_coro_wrapper(i: Interface, current_asyncio_task, wrapper_signature, 
                                             coro_worker_with_args: EntityArgsHolderExplicit, page_context: PageContextBase):
                coro_worker, args, kwargs = coro_worker_with_args.entity_args_kwargs()
                i.log.debug(f'awaitable_coro_wrapper - start: {entity_name(coro_worker)}')
                await apretent_to_be_asyncio_coro(i, current_asyncio_task)
                ClientHandlers.check_args_factory(wrapper_signature, args, kwargs)
                PageItems.check_args_factory(wrapper_signature, args, kwargs)
                # await i(Instance, InstanceRequest().wait(CS_PREPARED_FLAG))
                await page_context._ainit(i)
                i.log.debug(f'awaitable_coro_wrapper - ready: {entity_name(coro_worker)}')
                return await coro_worker(i, *args, **kwargs)
            
            coro_wrapper = awaitable_coro_wrapper
        elif CoroType.greenlet == coro_worker_type:
            def greenlet_coro_wrapper(i: Interface, current_asyncio_task, wrapper_signature, 
                                      coro_worker_with_args: EntityArgsHolderExplicit, page_context: PageContextBase):
                coro_worker, args, kwargs = coro_worker_with_args.entity_args_kwargs()
                i.log.debug(f'greenlet_coro_wrapper - start: {entity_name(coro_worker)}')
                pretent_to_be_asyncio_coro(i, current_asyncio_task)
                ClientHandlers.check_args_factory(wrapper_signature, args, kwargs)
                PageItems.check_args_factory(wrapper_signature, args, kwargs)
                # i(Instance, InstanceRequest().wait(CS_PREPARED_FLAG))
                page_context._init(i)
                i.log.debug(f'greenlet_coro_wrapper - ready: {entity_name(coro_worker)}')
                return coro_worker(i, *args, **kwargs)
            
            coro_wrapper = greenlet_coro_wrapper
        else:
            raise TypeError(f'{coro_worker} is neither an awaitable nor a greenlet')

        current_asyncio_task = asyncio.current_task()
        # args_kwargs: Optional[Tuple[Tuple, Dict]] = dict().get('args_kwargs', None)
        # if args_kwargs is None:
        #     args_kwargs = (tuple(), dict())
        
        args = tuple()
        kwargs = dict()
        coro_worker_param_names_set: Set[str] = set(inspect.signature(coro_worker).parameters.keys())
        if 'client' in coro_worker_param_names_set:
            kwargs['client'] = client

        if 'request' in coro_worker_param_names_set:
            kwargs['request'] = request
        
        translation_params: Set[str] = known_translation_param_names & coro_worker_param_names_set
        if translation_params:
            for translation_param in translation_params:
                kwargs[translation_param] = translatable_text_element
        
        page_context_params: Set[str] = known_page_context_param_names & coro_worker_param_names_set
        if page_context_params:
            page_class = page_class_0
            if isclass(page_class):
                if issubclass(page_class, PageContextBase):
                    page_context: PageContextBase = page_class(client, request, translatable_text_element, current_asyncio_task)
                else:
                    raise TypeError(f'{page_class} is not a subclass of PageContextBase')
            elif isinstance(page_class, EntityArgsHolder):
                page_class = cast(EntityArgsHolder, page_class)
                page_class, page_class_own_args, page_class_own_kwargs = page_class.entity_args_kwargs()
                page_class_own_kwargs.update({
                    'client': client,
                    'request': request,
                    '_t': translatable_text_element,
                    'current_asyncio_task': current_asyncio_task,
                })
                page_context = page_class(*page_class_own_args, **page_class_own_kwargs)
            else:
                raise TypeError(f'{page_class} is not a subclass of PageContextBase nor an EntityArgsHolder')

            for page_context_param in page_context_params:
                kwargs[page_context_param] = page_context

        return await await_coro_prim(coro_wrapper, current_asyncio_task, wrapper_signature, 
                                     EntityArgsHolderExplicit(coro_worker, args, kwargs), page_context)
        
    wrapper_sign: Signature = signature(wrapper)
    coro_worker_sign: Signature = signature(coro_worker)
    # wrapper_signature = coro_worker_sign.replace(parameters=tuple(coro_worker_sign.parameters.values())[1:], return_annotation=coro_worker_sign.return_annotation)
    wrapper_signature = coro_worker_sign.replace(parameters=wrapper_sign.parameters.values(), return_annotation=coro_worker_sign.return_annotation)
    wrapper.__signature__ = wrapper_signature
    
    return wrapper


nicegui_page_async_coro = partial(nicegui_page_async_coro_impl, None)


async_page = nicegui_page_async_coro
apage = nicegui_page_async_coro


def nicegui_page_class_async_coro(page_class: Optional[Union[PageContextBase, EntityArgsHolder]] = PageContextBase):
    return partial(nicegui_page_async_coro_impl, page_class)


async_page_class = nicegui_page_class_async_coro
apage_class = nicegui_page_class_async_coro


async def init_cs(is_fast_loop: bool = True, main_coro: AnyWorker = None, app_args_kwargs = None):
    async def coro(i: Interface, main_coro: AnyWorker = None, app_args_kwargs = None):
        set_primary_coro_scheduler(i._loop)
        app.cs = i._loop
        app.cs_initiated = False
        app_name: str = app_args_kwargs[1]['app_name']
        app_name_for_fs: str = app_args_kwargs[1]['app_name_for_fs']
        app_version: str = app_args_kwargs[1]['app_version']
        app_version_str: str = app_args_kwargs[1]['app_version_str']

        await i(Instance, InstanceRequest().set('app_name', app_name))
        await i(Instance, InstanceRequest().set('app_name_for_fs', app_name_for_fs))
        await i(Instance, InstanceRequest().set('app_version', app_version))
        await i(Instance, InstanceRequest().set('app_version_str', app_version_str))

        i.log.setLevel(logging.INFO)
        await i(LogRequest().sync())
        i.log.debug('cengal.nicegui.init_cs - start')
        await i(AsyncioLoop, AsyncioLoopRequest().inherit_surrounding_loop())
        await i(AsyncioLoop, AsyncioLoopRequest().turn_on_loops_intercommunication(True))
        await i(ShutdownOnKeyboardInterrupt)
        await i(Instance, InstanceRequest().set('cengal_nicegui__app', app))
        await i(Instance, InstanceRequest().set('cengal_nicegui__app_args_kwargs', app_args_kwargs))

        db_env_info: EnvInfo = await i(DbRequest().open_db_environment(db_env_id, None, False, max_dbs=20))
        await i(DbRequest(env_id=db_env_id).open_databases({
            'logged_in_clients',
            'user_clients',
            'logged_in_sessions',
            'user_sessions',
            'ip_geolocation',
            'user_sign_in_info',
            'credentials_by_user_id',
            'user_sign_up_info',
        }))
        await i(Instance, InstanceRequest().set('cengal_nicegui__db_env_info', db_env_info))
        await i(Instance, InstanceRequest().set('cengal_nicegui__session_pages', dict()))
        await i(Instance, InstanceRequest().set('cengal_nicegui__user_clients_lock', Lock(f'cengal_nicegui__user_clients_lock__{uuid4()}')))
        await i(Instance, InstanceRequest().set('cengal_nicegui__user_sessions_lock', Lock(f'cengal_nicegui__user_sessions_lock__{uuid4()}')))

        # await init_translation(i, app_args_kwargs)
        if main_coro is not None:
            await i(PutCoro, main_coro, app_args_kwargs)
        
        app.cs_initiated = True
        i.log.debug('cengal.nicegui.init_cs - ready')
        await i(Instance, InstanceRequest().set(CS_PREPARED_FLAG, True))
        try:
            await i(AsyncEventBus, AsyncEventBusRequest().wait(DESTROY_CS_EVENT))
        finally:
            await i(DbRequest().close_db_environment(db_env_id))
            i.log.debug('cengal.nicegui.init_cs - shutting down loop')
            await i(ShutdownLoop)

    if is_fast_loop:
        await arun_in_fast_loop(coro, main_coro, app_args_kwargs)
    else:
        await arun_in_loop(coro, main_coro, app_args_kwargs)


# # Async on_destroy handling is currently (08 Feb 2023) broken in NiceGUI: 
# #     they should await for handlers instead of current tasks creation approach
# async def destroy_cs():
#     async def coro(i: Interface):
#         await i(AsyncEventBus, AsyncEventBusRequest().send_event(DESTROY_CS_EVENT, None))
#         await i(Sleep, CS_DESTROY_TIMEOUT)

#     await await_coro_prim(coro)


# Async on_destroy handling is currently (08 Feb 2023) broken in NiceGUI: 
#     they should await for handlers instead of current tasks creation approach
def destroy_cs():
    cs: CoroSchedulerType = get_available_coro_scheduler()
    cs_destroy_timeouted: bool = False
    if cs and (not cs._destroyed):
        async def coro(i: Interface):
            await i(AsyncEventBus, AsyncEventBusRequest().send_event(DESTROY_CS_EVENT, None))

        put_coro_to(get_interface_and_loop_with_explicit_loop(cs), coro)
        start_time = perf_counter()
        while cs.iteration():
            if (perf_counter() - start_time) >= CS_DESTROY_TIMEOUT:
                cs_destroy_timeouted = True
                break

    if cs_destroy_timeouted:
        warnings.warn(f'destroy_cs - not finished within timeout of {CS_DESTROY_TIMEOUT} sec.')


@asyncio_coro
async def on_disconnect_handler(i: Interface, client: Client):
    client_id = client.id
    if hasattr(client, 'session_id'):
        session_id = client.session_id
    else:
        session_id = None
    
    if session_id is None:
        translatable_text_element: NiceguiTranslatableTextElement = translatable_text_element_per_client.pop(client_id, None)
    else:
        translatable_text_element = translatable_text_element_per_session.pop(session_id, None)
    
    if hasattr(client, 'translatable_text_element'):
        if translatable_text_element is None:
            translatable_text_element = client.translatable_text_element
        
        delattr(client, 'translatable_text_element')

    if translatable_text_element is not None:
        await translatable_text_element.aremove_on_lang_changed_handler()
    
    if session_id is not None:
        session_clients[session_id].discard(client_id)
        if not session_clients[session_id]:
            session_clients.pop(session_id, None)


async def init_translation(i: Interface, app_args_kwargs: Tuple[Tuple, Dict]):
    await i(ShutdownOnKeyboardInterrupt)
    app_dir_path: AppDirPath = await i(Instance, InstanceRequest().get(AppDirPath))
    app_data_dir_path_type: str = await i(Instance, InstanceRequest().get('app_data_dir_path_type'))
    app_name_for_fs: str = await i(Instance, InstanceRequest().get('app_name_for_fs'))
    app_name_for_fs = app_name_for_fs if app_name_for_fs else app_args_kwargs[1]['app_name_for_fs']
    app_data_dir_path_rel: RelativePath = RelativePath(app_dir_path(app_data_dir_path_type, app_name_for_fs))
    global text_translator
    global translation_language_mapper
    text_translator, translation_language_mapper = setup_translation(app_data_dir_path_rel('text_dictionary.json'))
    await i(Instance, InstanceRequest().set('cengal_nicegui__text_translator', text_translator))
    await i(Instance, InstanceRequest().set('cengal_nicegui__translation_language_mapper', translation_language_mapper))


def run(*,
        host: str = '0.0.0.0',
        port_or_range: Union[int, slice, Tuple[int, int]] = 8080,
        main_coro: AnyWorker = None,
        is_fast_loop: bool = True,
        app_name: str = str(),
        app_name_for_fs: str = str(),
        app_version: Tuple[int, int, int, Union[int, str]] = tuple(),
        app_version_str: str = str(),
        **kwargs
    ) -> None:
    """Prepares and starts NiceGUI. Saves initial args and kwargs parameters (as well as determined free port) into a tuple (Tuple[Tuple, Dict]) within an Instance service awailable by a 'cengal_nicegui__app_args_kwargs' string key 

    Args:
        host (str, optional): _description_. Defaults to '0.0.0.0'.
        port_or_range (Union[int, slice, Tuple[int, int]], optional): _description_. Defaults to 8080.
        title (str, optional): _description_. Defaults to 'NiceGUI'.
        viewport (str, optional): _description_. Defaults to 'width=device-width, initial-scale=1'.
        favicon (Optional[str], optional): _description_. Defaults to None.
        dark (Optional[bool], optional): _description_. Defaults to False.
        binding_refresh_interval (float, optional): _description_. Defaults to 0.1.
        show (bool, optional): _description_. Defaults to True.
        reload (bool, optional): _description_. Defaults to True.
        uvicorn_logging_level (str, optional): _description_. Defaults to 'warning'.
        uvicorn_reload_dirs (str, optional): _description_. Defaults to '.'.
        uvicorn_reload_includes (str, optional): _description_. Defaults to '*.py'.
        uvicorn_reload_excludes (str, optional): _description_. Defaults to '.*, .py[cod], .sw.*, ~*'.
        tailwind (bool, optional): _description_. Defaults to True.
        is_fast_loop (bool, optional): _description_. Defaults to True.
        main_coro (AnyWorker, optional): _description_. Defaults to None.
    """
    port = simple_port_search(host, port_or_range)
    app_args_kwargs: Tuple[Tuple, Dict] = args_kwargs(
        host=host, 
        port_or_range=port_or_range, 
        port=port, 
        main_coro=main_coro,
        is_fast_loop=is_fast_loop,
        app_name=app_name,
        app_name_for_fs=app_name_for_fs,
        app_version=app_version,
        app_version_str=app_version_str,
        **kwargs,
    )
    run_in_loop(init_translation, app_args_kwargs)
    app.on_startup(init_cs(is_fast_loop, main_coro, app_args_kwargs))
    app.on_shutdown(destroy_cs)
    # app.on_connect(on_connect_handler)
    app.on_disconnect(on_disconnect_handler)
    ui.run(
        host=host,
        port=port,
        **kwargs,
        )



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = ['PutSingleCoroParams', 'PSCP', 'PutCoroList', 'put_coro_list_to', 'try_put_coro_list_to', 'aput_coro_list_to', 'atry_put_coro_list_to', 'put_coro_list', 'try_put_coro_list', 'aput_coro_list', 'atry_put_coro_list']


from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import PutCoro, Task
from cengal.parallel_execution.coroutines.coro_standard_services_internal_lib.service_with_a_direct_request import *
from cengal.code_flow_control.smart_values import ValueExistence
from cengal.introspection.inspect import get_exception, get_exception_tripple
from cengal.code_flow_control.args_manager import EntityArgsHolder
from typing import Sequence, Tuple, List, Optional, Any, cast, Dict, Union


class PutSingleCoroParams:
    def __init__(self, coro_worker: AnyWorker, *args, **kwargs) -> None:
        self.coro_worker: AnyWorker = coro_worker
        self.args: Tuple = args
        self.kwargs: Dict = kwargs
    
    def __call__(self) -> Any:
        return self.coro_worker, self.args, self.kwargs


PSCP = PutSingleCoroParams


class PutCoroList(TypedService[List[Tuple[Optional[CoroID], Optional[Exception]]]], ServiceWithADirectRequestMixin):
    def __init__(self, loop: CoroSchedulerType):
        super(PutCoroList, self).__init__(loop)
        self.direct_requests: List[Tuple] = list()

    def single_task_registration_or_immediate_processing(
            self, coro_list: Sequence[PutSingleCoroParams], tasks: bool = False
    ) -> Tuple[bool, Union[List[CoroID], List[Task]], Any]:
        results = list()
        try:
            put_coro: PutCoro = self._loop.get_service_instance(PutCoro)
            caller_coro_id: CoroID = self.current_caller_coro_info.coro_id
            for request in coro_list:
                result = None
                exception = None
                try:
                    coro_worker, args, kwargs = request()
                    if tasks:
                        result = put_coro.put_task_from_other_service(caller_coro_id, coro_worker, *args, **kwargs)
                    else:
                        coro = put_coro.put_from_other_service(caller_coro_id, coro_worker, *args, **kwargs)
                        result = coro.coro_id
                except:
                    exception = get_exception()

                results.append((result, exception))
        except:
            return True, results, get_exception()

        return True, results, None

    def full_processing_iteration(self):
        put_coro: PutCoro = self._loop.get_service_instance(PutCoro)
        direct_requests_buff = self.direct_requests
        self.direct_requests = type(direct_requests_buff)()
        for coro_list in direct_requests_buff:
            for request in coro_list:
                coro_worker, args, kwargs = request()
                try:
                    coro = put_coro.put_root_from_other_service(coro_worker, *args, **kwargs)
                except:
                    ex_type, exception, tracback = get_exception_tripple()
                    if __debug__: dlog(ex_type, exception, tracback)
                    raise

        self.make_dead()
    
    def _add_direct_request(self, coro_list: Sequence[PutSingleCoroParams]) -> ValueExistence[None]:
        self.direct_requests.append(coro_list)
        self.make_live()
        return (False, None)

    def in_work(self) -> bool:
        result = bool(self.direct_requests)
        return self.thrifty_in_work(result)


def put_coro_list_to(context: Tuple[Optional[CoroSchedulerType], Optional[Interface], bool], coro_list: Sequence[PutSingleCoroParams]) -> ValueExistence[CoroID]:
    return make_request_to_service_with_context(context, PutCoroList, coro_list)


def try_put_coro_list_to(context: Tuple[Optional[CoroSchedulerType], Optional[Interface], bool], coro_list: Sequence[PutSingleCoroParams]) -> ValueExistence[Optional[CoroID]]:
    return try_make_request_to_service_with_context(context, PutCoroList, coro_list)


async def aput_coro_list_to(context: Tuple[Optional[CoroSchedulerType], Optional[Interface], bool], coro_list: Sequence[PutSingleCoroParams]) -> ValueExistence[CoroID]:
    return await amake_request_to_service_with_context(context, PutCoroList, coro_list)


async def atry_put_coro_list_to(context: Tuple[Optional[CoroSchedulerType], Optional[Interface], bool], coro_list: Sequence[PutSingleCoroParams]) -> ValueExistence[Optional[CoroID]]:
    return await atry_make_request_to_service_with_context(context, PutCoroList, coro_list)


def put_coro_list(coro_list: Sequence[PutSingleCoroParams]) -> ValueExistence[CoroID]:
    return make_request_to_service(PutCoroList, coro_list)


def try_put_coro_list(coro_list: Sequence[PutSingleCoroParams]) -> ValueExistence[Optional[CoroID]]:
    return try_make_request_to_service(PutCoroList, coro_list)


async def aput_coro_list(coro_list: Sequence[PutSingleCoroParams]) -> ValueExistence[CoroID]:
    return await amake_request_to_service(PutCoroList, coro_list)


async def atry_put_coro_list(coro_list: Sequence[PutSingleCoroParams]) -> ValueExistence[Optional[CoroID]]:
    return await atry_make_request_to_service(PutCoroList, coro_list)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

from cengal.math.numbers import RationalNumber
from cengal.time_management.repeat_for_a_time import Tracer
from math import sqrt, pi, factorial

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"

class Ellipse:
    """
    https://www.youtube.com/watch?v=5nW3nJhBHL0
    http://www.ebyte.it/library/docs/math05a/EllipsePerimeterApprox05.html
    """
    def __init__(self, a: RationalNumber, b: RationalNumber):
        self.a = a
        self.b = b
    
    def y(self, x: RationalNumber) -> RationalNumber:
        return self.b * sqrt(1 - (x / self.a)**2)
    
    @property
    def area(self):
        return pi * self.a * self.b
    
    @property
    def h(self):
        return ((self.a - self.b)**2) / ((self.a + self.b)**2)
    
    @property
    def e(self):
        return sqrt(self.a**2 - self.b**2) / self.a
    
    @property
    def c(self):
        return self.a * self.e
    
    @property
    def perimeter__kepler(self):
        return 2 * pi * sqrt(self.a * self.b)
    
    @property
    def perimeter__naive(self):
        return 2 * pi * ((self.a + self.b) / 2)
    
    @property
    def perimeter__euler(self):
        return 2 * pi * sqrt((self.a**2 - self.b**2) / 2)
    
    @property
    def perimeter__matt_parker__lazy(self):
        """Computationly efficient. Not more that 6.1% deviation (less than 5% for `1 / 75` ellipses)

        Returns:
            _type_: _description_
        """
        a = max(self.a, self.b)
        b = min(self.a, self.b)
        return pi * ((6 * a) / 5 + (3 * b) / 4)
    
    @property
    def perimeter__best_of__euler__and__matt_parker__lazy(self):
        a = max(self.a, self.b)
        b = min(self.a, self.b)
        if (a / b) >= 0.8:
            return self.perimeter__matt_parker__lazy
        else:
            return self.perimeter__euler
    
    @property
    def perimeter__ramanujan_1(self):
        return pi * (3 * (self.a + self.b) - sqrt((3 * self.a + self.b) * (self.a + 3 * self.b)))
    
    @property
    def perimeter__matt_parker__precise(self):
        a = max(self.a, self.b)
        b = min(self.a, self.b)
        return pi * ((53 * a) / 3 + (717 * b) / 35 - sqrt(269 * a ** 2 + 667 * a * b + 371 * b ** 2))
    
    @property
    def perimeter__best_of__ramanujan_1__and__matt_parker__precise(self):
        a = max(self.a, self.b)
        b = min(self.a, self.b)
        if (a / b) >= 2.4:
            return self.perimeter__matt_parker__precise
        else:
            return self.perimeter__ramanujan_1
    
    @property
    def perimeter__ramanujan_2(self):
        return pi * (self.a + self.b) * (1 + (3 * self.h) / (10 + sqrt(4 - 3 * self.h)))
    
    def perimeter__infinite_sum__time_lim(self, desired_approximation_time: float, return_iterations_num: bool = False):
        e = self.e
        first_multiplier = 2 * self.a * pi
        second_multiplier = 1
        tr = Tracer(desired_approximation_time)
        while tr.iter():
            i = tr.total_number_of_iterations_made
            second_multiplier -= ((factorial(2 * i) ** 2) / ((2 ** i * factorial(i)) ** 4)) * ((e ** (2 * i)) / (2 * i - 1))

        if return_iterations_num:
            return first_multiplier * second_multiplier, tr.total_number_of_iterations_made
        else:
            return first_multiplier * second_multiplier
    
    def perimeter__infinite_sum__iter_lim(self, iterations_num: int):
        e = self.e
        first_multiplier = 2 * self.a * pi
        second_multiplier = 1
        for i in range(1, iterations_num + 1):
            second_multiplier -= ((factorial(2 * i) ** 2) / ((2 ** i * factorial(i)) ** 4)) * ((e ** (2 * i)) / (2 * i - 1))

        return first_multiplier * second_multiplier
    
    @staticmethod
    def from_r1_r2(r1: RationalNumber, r2: RationalNumber):
        return Ellipse((r1 + r2) / 2, 1)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import namedtuple
from dis import Instruction, dis, get_instructions, _get_code_object, code_info, show_code, findlabels, _unpack_opargs
from opcode import hasjrel, hasjabs, opname, opmap, HAVE_ARGUMENT, EXTENDED_ARG
from cengal.system import PYTHON_VERSION_INT
from enum import Enum
from typing import Optional, Callable, List, Tuple, Dict, Set, Union, Any, Sequence
from types import CodeType
from copy import copy
from cengal.entities.copyable import CopyableMixin
from cengal.data_manipulation.front_triggerable_variable import FrontTriggerableVariableType, FrontTriggerableVariable
import sys


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


BytecodeSequence = Union[bytes, Sequence[Union[int, bytes]]]


def patch_function(func, co_code):
    fn_code = func.__code__
    func.__code__ = CodeType(
        fn_code.co_argcount,
        fn_code.co_kwonlyargcount,
        fn_code.co_nlocals,
        fn_code.co_stacksize,
        fn_code.co_flags,
        co_code,
        fn_code.co_consts,
        fn_code.co_names,
        fn_code.co_varnames,
        fn_code.co_filename,
        fn_code.co_name,
        fn_code.co_firstlineno,
        fn_code.co_lnotab,
        fn_code.co_freevars,
        fn_code.co_cellvars,
    )


CodeParamNames = namedtuple("CodeParamNames", "positional positional_only keyword_only")


def code_param_names(code) -> CodeParamNames:
    pos_count = code.co_argcount
    arg_names = code.co_varnames
    positional = arg_names[:pos_count]
    posonly_count = code.co_posonlyargcount
    positional_only = arg_names[:posonly_count]
    keyword_only_count = code.co_kwonlyargcount
    keyword_only = arg_names[pos_count:pos_count + keyword_only_count]
    return CodeParamNames(positional, positional_only, keyword_only)


def code_name(code: CodeType) -> str:
    return code.co_name


if sys.version_info >= (3, 11):
    def code_qualname(code: CodeType) -> str:
        return code.co_qualname


def get_code(x = None) -> CodeType:
    if x is None:
        return
    
    if isinstance(x, CodeType):
        return x
    
    # Extract functions from methods.
    if hasattr(x, '__func__'):
        x = x.__func__
    # Extract compiled code objects from...
    if hasattr(x, '__code__'):  # ...a function, or
        x = x.__code__
    elif hasattr(x, 'gi_code'):  #...a generator object, or
        x = x.gi_code
    elif hasattr(x, 'ag_code'):  #...an asynchronous generator object, or
        x = x.ag_code
    elif hasattr(x, 'cr_code'):  #...a coroutine, or
        x = x.cr_code
    elif hasattr(x, 'f_code'):  #...a frame.
        x = x.f_code
    # else:
    #     raise TypeError(f'Expected a code object or an entity with code, but got {type(x)}')
    
    return x


def has_code(x = None) -> bool:
    if x is None:
        return False
    
    if isinstance(x, CodeType):
        return True
    
    # Extract functions from methods.
    if hasattr(x, '__func__'):
        return True
    # Extract compiled code objects from...
    if hasattr(x, '__code__'):  # ...a function, or
        return True
    elif hasattr(x, 'gi_code'):  #...a generator object, or
        return True
    elif hasattr(x, 'ag_code'):  #...an asynchronous generator object, or
        return True
    elif hasattr(x, 'cr_code'):  #...a coroutine, or
        return True
    elif hasattr(x, 'f_code'):  #...a frame.
        return True
    # else:
    #     raise TypeError(f'Expected a code object or an entity with code, but got {type(x)}')
    
    return False


class CodeTypeEnum(Enum):
    class_or_module = 0
    code_object = 1
    raw_bytecode = 2
    source_code = 3
    unknown = 4


def code_type(x=None) -> Optional[CodeTypeEnum]:
    """_summary_

    Args:
        x (_type_, optional): result of get_code() function. Defaults to None.

    Returns:
        Optional[CodeTypeEnum]: _description_
    """
    if x is None:
        return
    
    # Perform the disassembly.
    if hasattr(x, '__dict__'):  # Class or module
        return CodeTypeEnum.class_or_module
    elif hasattr(x, 'co_code'): # Code object
        return CodeTypeEnum.code_object
    elif isinstance(x, (bytes, bytearray)): # Raw bytecode
        return CodeTypeEnum.raw_bytecode
    elif isinstance(x, str):    # Source code
        return CodeTypeEnum.source_code
    else:
        return CodeTypeEnum.unknown


def set_code(x, code: CodeType):
    # Extract functions from methods.
    if hasattr(x, '__func__'):
        x.__func__ = code
    # Extract compiled code objects from...
    if hasattr(x, '__code__'):  # ...a function, or
        x.__code__ = code
    elif hasattr(x, 'gi_code'):  #...a generator object, or
        x.gi_code = code
    elif hasattr(x, 'ag_code'):  #...an asynchronous generator object, or
        x.ag_code = code
    elif hasattr(x, 'cr_code'):  #...a coroutine.
        x.cr_code = code
    
    return x


if sys.version_info < (3, 11):
# if (3, 11) > PYTHON_VERSION_INT[:1]:
    def modify_code(original_code: CodeType, co_code, co_consts, co_names, co_varnames):
        co_nlocals = len(co_varnames)
        return CodeType(
            original_code.co_argcount,
            original_code.co_posonlyargcount,
            original_code.co_kwonlyargcount,
            co_nlocals,
            original_code.co_stacksize,
            original_code.co_flags,
            co_code,
            tuple(co_consts),
            tuple(co_names),
            tuple(co_varnames),
            original_code.co_filename,
            original_code.co_name,
            original_code.co_firstlineno,
            original_code.co_lnotab,
            original_code.co_freevars,
            original_code.co_cellvars,
        )


    # def unpack_opargs_original(code, denormalize_values: bool = False):
    #     ftv: FrontTriggerableVariable = FrontTriggerableVariable(FrontTriggerableVariableType.equal, EXTENDED_ARG)
    #     extended_arg = 0
    #     real_op_index: Optional[int] = None
    #     real_byte_index: Optional[int] = None
    #     need_to_clear_real_data: bool = False

    #     op_index = -1
    #     for i in range(0, len(code), 2):
    #         op_index += 1
    #         op = code[i]
    #         if op >= HAVE_ARGUMENT:
    #             ftv_result: Optional[bool] = ftv(op)
    #             if ftv_result is True:
    #                 real_op_index = op_index
    #                 real_byte_index = i
    #             elif ftv_result is False:
    #                 need_to_clear_real_data = True

    #             arg = code[i+1] | extended_arg
    #             if denormalize_values:
    #                 extended_arg = (arg << 8) if op == EXTENDED_ARG else 0
    #         else:
    #             arg = None
            
    #         yield (op, arg, op_index, i, real_op_index, real_byte_index)
    #         if need_to_clear_real_data:
    #             real_op_index = None
    #             real_byte_index = None
    #             need_to_clear_real_data = False

    def unpack_opargs(code: BytecodeSequence, denormalize_values: bool = False):
        ftv: FrontTriggerableVariable = FrontTriggerableVariable(FrontTriggerableVariableType.equal, EXTENDED_ARG)
        extended_arg = 0
        real_op_index: Optional[int] = None
        real_offset: Optional[int] = None
        need_to_clear_real_data: bool = False

        op = None
        arg = None
        offset = -1
        op_index = -1
        i = -1
        for item in code:
            if isinstance(item, bytes):
                item = int.from_bytes(item, byteorder='little')
            
            i += 1
            if not i % 2:
                op = item
                op_index += 1
                offset = i
                continue
            
            if op >= HAVE_ARGUMENT:
                ftv_result: Optional[bool] = ftv(op)
                if ftv_result is True:
                    real_op_index = op_index
                    real_offset = offset
                elif ftv_result is False:
                    need_to_clear_real_data = True

                arg = item | extended_arg
                if denormalize_values:
                    extended_arg = (arg << 8) if op == EXTENDED_ARG else 0
            else:
                arg = None
            
            yield (op, arg, op_index, offset, real_op_index, real_offset)
            if need_to_clear_real_data:
                real_op_index = None
                real_offset = None
                need_to_clear_real_data = False

    def find_ops_with_labels(code: BytecodeSequence) -> Tuple[List[int], Dict[int, List[Tuple[int, int, int, int]]]]:
        labels: List[int] = list()
        op_by_label: Dict[int, List[Tuple[int, int, int, int]]] = dict()
        for op, arg, op_index, offset, real_op_index, real_offset in unpack_opargs(code, True):
            if arg is not None:
                if op in hasjrel:
                    label = offset + 2 + arg
                elif op in hasjabs:
                    label = arg
                else:
                    continue

                if label not in op_by_label:
                    labels.append(label)
                    op_by_label[label] = list()
                
                op_by_label[label].append((op, arg, op_index, offset, real_op_index, real_offset))

        return labels, op_by_label
    
    
    def op_index_to_arg(op_index: int) -> int:
        return op_index * 2
    

    def arg_to_op_index(arg: int) -> int:
        return arg // 2

elif sys.version_info == (3, 11):
    from dis import _is_backward_jump

    def find_ops_with_labels(code):
        labels = list()
        op_by_label = dict()
        index = -1
        for offset, op, arg in _unpack_opargs(code):
            index += 1
            if arg is not None:
                if op in hasjrel:
                    if _is_backward_jump(op):  # inefficient implementation of the _is_backward_jump() function. Need to use set or list instead of string manipulations
                        arg = -arg

                    label = offset + 2 + arg*2  # In 3.10: this is an instruction index - not a byte index. Need to be fixed and tested
                elif op in hasjabs:
                    label = arg*2
                else:
                    continue

                if label not in op_by_label:
                    labels.append(label)
                    op_by_label[label] = list()
                
                op_by_label[label].append((index, offset, op, arg))

        return labels, op_by_label
else:
    import warnings
    warnings.warn(f'Unsupported Python Version: {PYTHON_VERSION_INT}')
    # raise RuntimeError(f'Unsupported Python Version: {PYTHON_VERSION_INT}')


class OpSequenceOffsetMap(CopyableMixin):
    def __init__(self, op_num: int = 0):
        self.op_num: int = op_num
        self.range: Optional[slice] = slice(0, op_num) if op_num else None
        self.new_by_original: Dict[int, int] = dict()
        self.original_by_new: Dict[int, int] = dict()
        if op_num < 0:
            op_num = 0
        
        for index in range(op_num):
            # bindex = index * 2
            bindex = index
            self.new_by_original[bindex] = bindex
            self.original_by_new[bindex] = bindex
    
    def mapped_range(self) -> slice:
        return slice(min(self.original_by_new), max(self.original_by_new))

    
    def remove_slice(self, op_index: int, op_num: int, preserver_index_for_first_x_op: int = 0):
        if not op_num:
            return
        
        offset_change_start = op_index
        del_start = op_index
        del_stop = op_index + op_num

        new_by_original_buff = self.new_by_original
        self.new_by_original = type(self.new_by_original)()
        to_be_deleted: List[Tuple[int, int]] = list()
        for original, new in new_by_original_buff.items():
            if del_start <= new < del_stop:
                to_be_deleted.append((original, new))
        
        to_be_deleted.sort(key=lambda x: x[1])
        if preserver_index_for_first_x_op:
            to_be_deleted = to_be_deleted[preserver_index_for_first_x_op:]
        
        for original, new in to_be_deleted:
            del self.new_by_original[original]
        
        original_by_new_buff = self.original_by_new
        self.original_by_new = type(self.original_by_new)()
        to_be_deleted = list()
        for new, original in original_by_new_buff.items():
            if del_start <= new < del_stop:
                to_be_deleted.append((new, original))
        
        to_be_deleted.sort(key=lambda x: x[0])
        if preserver_index_for_first_x_op:
            ignored = to_be_deleted[:preserver_index_for_first_x_op]
            if ignored:
                last_ignored = ignored[-1]
                last_ignored_new = last_ignored[0]
                offset_change_start = last_ignored_new + 1
            
            to_be_deleted = to_be_deleted[preserver_index_for_first_x_op:]
        
        for new, original in to_be_deleted:
            del self.original_by_new[new]
        
        self.add_offset(-op_num, offset_change_start)
        self.op_num -= op_num
        self.range = slice(self.range.start, self.range.stop - op_num)
    
    def insert_slice(self, op_index: int, op_num: int, preserver_index_for_first_x_op: int = 0):
        if not op_num:
            return
        
        self.add_offset(op_num, op_index + preserver_index_for_first_x_op)
        self.op_num += op_num
        self.range = slice(self.range.start, self.range.stop + op_num)
    
    def insert_op_sequence_offset(self, op_index: int, op_sequence_offset_map: 'OpSequenceOffsetMap', insertion_id: Any, preserver_index_for_first_x_op: int = 0):
        op_num: int = op_sequence_offset_map.op_num
        if not op_num:
            return
        
        sequence_range: slice = op_sequence_offset_map.range
        offset = op_index - sequence_range.start
        self.insert_slice(op_index, op_num, preserver_index_for_first_x_op)
        for new, original in op_sequence_offset_map.original_by_new.items():
            self.original_by_new[new + offset] = (insertion_id, new)
    
    def add_offset(self, op_offset: int, op_start: int = 0):
        if not op_offset:
            return
        
        new_by_original_buff = self.new_by_original
        self.new_by_original = type(self.new_by_original)()
        for original, new in new_by_original_buff.items():
            if new >= op_start:
                new += op_offset
            
            self.new_by_original[original] = new
        
        original_by_new_buff = self.original_by_new
        self.original_by_new = type(self.original_by_new)()
        for new, original in original_by_new_buff.items():
            if new >= op_start:
                new += op_offset
            
            self.original_by_new[new] = original
    
    def shift(self, op_offset: int):
        self.add_offset(op_offset)
        self.range = slice(self.range.start + op_offset, self.range.stop + op_offset)
    
    def shift_to_absolute(self, op_index: int):
        op_offset: int = op_index - self.range.start
        self.add_offset(op_offset)
        self.range = slice(self.range.start + op_offset, self.range.stop + op_offset)
    
    def update(self, op_sequence_offset_map: 'OpSequenceOffsetMap', op_offset: int = 0):
        if op_offset:
            op_sequence_offset_map = op_sequence_offset_map.copy()
            op_sequence_offset_map.add_offset(op_offset)
        
        self.new_by_original.update(op_sequence_offset_map.new_by_original)
        self.original_by_new.update(op_sequence_offset_map.original_by_new)
    
    def __getitem__(self, index: Union[int, slice]) -> 'OpSequenceOffsetMap':
        if isinstance(index, int):
            index = slice(index, index + 1)
        
        op_num = index.stop - index.start

        new_by_original_buff = self.new_by_original
        new_by_original = type(new_by_original_buff)()
        for original, new in new_by_original_buff.items():
            if index.start <= new < index.stop:
                new_by_original[original] = new
        
        original_by_new_buff = self.original_by_new
        original_by_new = type(original_by_new_buff)()
        for new, original in original_by_new_buff.items():
            if index.start <= new < index.stop:
                original_by_new[new] = original

        result = OpSequenceOffsetMap()
        result.op_num = op_num
        result.range = index
        result.new_by_original = new_by_original
        result.original_by_new = original_by_new
        return result

    def copy(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__['new_by_original'] = copy(self.new_by_original)
        result.__dict__['original_by_new'] = copy(self.original_by_new)
        result.__dict__['op_num'] = self.op_num
        result.__dict__['range'] = copy(self.range)
        return result


def opcode(name: str) -> int:
    name = name.upper()
    return opmap[name]


def opcode_name(opcode: int) -> str:
    return opname[opcode]


class OpSequence:
    extended_arg_opcode_int = opcode('EXTENDED_ARG')
    extended_arg_opcode_byte = extended_arg_opcode_int.to_bytes(1, byteorder='little')

    def __init__(self, op_sequence: Optional[List[Tuple[int, int, int, int, int, int]]] = None):
        self.op_sequence: List[Tuple[int, int, int, int, int, int]] = list() if op_sequence is None else op_sequence
        self.op_sequence_offset_map: OpSequenceOffsetMap = OpSequenceOffsetMap(len(self.op_sequence))
    
    def __len__(self):
        return len(self.op_sequence)
    
    def read_slice(self, place: slice) -> 'OpSequence':
        result = OpSequence()
        result.op_sequence = self.op_sequence[place]
        result.op_sequence_offset_map = self.op_sequence_offset_map[place]
        result.op_sequence_offset_map.shift_to_absolute(0)
        return result
    
    def remove_slice(self, place: slice, preserver_index_for_first_x_op: int = 0):
        self.op_sequence[place] = []
        self.op_sequence_offset_map.remove_slice(place.start, place.stop - place.start, preserver_index_for_first_x_op)
    
    def insert_op_sequence(self, index: int, op_sequence: 'OpSequence', preserver_index_for_first_x_op: int = 0):
        self.op_sequence[index: index] = op_sequence.op_sequence
        op_sequence.op_sequence_offset_map.shift_to_absolute(index)
        self.op_sequence_offset_map.insert_op_sequence_offset(index, op_sequence.op_sequence_offset_map, op_sequence, preserver_index_for_first_x_op)
        return op_sequence
    
    def normalize_instructions_arg(self):
        index = -1
        while True:
            index += 1
            if index >= len(self.op_sequence):
                break

            op, arg, op_index, offset, real_op_index, real_offset = self.op_sequence[index]
            if arg is not None:
                if arg > 255:
                    arg = arg_to_bytes(arg)
                    if real_op_index is not None:
                        real_op_index_delta = real_op_index - op_index
                        real_op_index = index - real_op_index_delta
                        slice_to_delete = slice(real_op_index, index)
                        self.remove_slice(slice_to_delete, 1)
                        index = real_op_index

                    extended_arg: bytes = arg[1:]
                    extended_arg_len = len(extended_arg)
                    arg = arg[0]
                    self.op_sequence[index] = (op, arg, index + extended_arg_len, (index + extended_arg_len) * 2, index, index * 2)
                    op_sub_sequence_instructions = list()
                    for extended_arg_int in reversed(extended_arg):
                        op_sub_sequence_instructions.append(make_instruction('EXTENDED_ARG', extended_arg_int))
                    
                    sub_op_sequence: 'OpSequence' = OpSequence.from_instructions_fast(op_sub_sequence_instructions)
                    self.insert_op_sequence(index, sub_op_sequence, 1)
                    index += len(sub_op_sequence.op_sequence)
    
    def denormalize_instructions_arg(self):
        raise NotImplementedError
    
    def find_op(self, op_index: int) -> Optional[slice]:
        start_index = None
        end_index = None

        index = op_index - 1
        while True:
            index += 1
            if index >= len(self.op_sequence):
                break

            current_op, current_arg, _, _, _, _ = self.op_sequence[index]
            if current_op == OpSequence.extended_arg_opcode_int:
                continue
            else:
                end_index = index + 1
                break
        
        previous_op_is_extended_arg = None
        index = op_index
        while True:
            index -= 1
            if index < 0:
                if previous_op_is_extended_arg:
                    start_index = 0
                
                break
            
            current_op, current_arg, _, _, _, _ = self.op_sequence[index]
            if current_op == OpSequence.extended_arg_opcode_int:
                previous_op_is_extended_arg = True
                continue
            else:
                previous_op_is_extended_arg = False
                start_index = index + 1
                break
        
        if (start_index is None) or (end_index is None):
            return None
        
        return slice(start_index, end_index)
    
    def get_arg(self, op_index: Union[int, slice]) -> bytes:
        if isinstance(op_index, int):
            op_index = self.find_op(op_index)
        
        if (op_index is None) or ((op_index.stop - op_index.start) < 1):
            raise RuntimeError('OP not found by index')
        
        op_slice = self.op_sequence[op_index]
        if (op_index.stop - op_index.start) > 1:
            arg_list = list()
            for op, arg, _, _, _, _ in reversed(op_slice):
                arg_list.append(arg)
            
            return arg_to_int(bytes(arg_list))
        else:
            return op_slice[0][1]
    
    def set_arg(self, op_index: Union[int, slice], new_arg: Union[None, int, bytes]):
        if isinstance(op_index, int):
            op_index = self.find_op(op_index)
        
        op_index_len = op_index.stop - op_index.start
        if (op_index is None) or (op_index_len < 1):
            raise RuntimeError('OP not found by index')
        
        new_arg = arg_to_int(new_arg)
        current_op, current_arg, current_op_index, current_offset, current_real_op_index, current_real_offset = self.op_sequence[op_index.stop - 1]
        sub_op_sequence_bytes = [current_op, new_arg]
        sub_op_sequence: OpSequence = OpSequence(list(unpack_opargs(sub_op_sequence_bytes)))
        sub_op_sequence.normalize_instructions_arg()
        real_arg = sub_op_sequence.op_sequence[-1][1]
        sub_op_sequence_len = len(sub_op_sequence.op_sequence)
        extended_args_len = sub_op_sequence_len - 1
        if extended_args_len:
            current_real_op_index = op_index.start
            current_real_offset = current_real_op_index * 2
            current_op_index = current_real_op_index + extended_args_len
            current_offset = current_op_index * 2
        else:
            current_op_index = op_index.start
            current_offset = current_op_index * 2
            current_real_op_index = None
            current_real_offset = None

        self.op_sequence[op_index.stop - 1] = (current_op, real_arg, current_op_index, current_offset, current_real_op_index, current_real_offset)

        if sub_op_sequence_len > 1:
            sub_op_sequence.remove_slice(slice(sub_op_sequence_len - 1, sub_op_sequence_len))
            if op_index_len > 1:
                self.remove_slice(slice(op_index.start, op_index.stop - 1), 1)
            
            self.insert_op_sequence(op_index.start, sub_op_sequence, 1)

    def fix_labels(self, op_by_label: Dict[int, List[int]]):
        labels = list(op_by_label.keys())
        labels.sort()
        for label, data in op_by_label.items():
            data.sort()
        
        for label in labels:
            labeled_op_index_new = self.op_sequence_offset_map.new_by_original[label]
            label_data = op_by_label[label]
            while True:
                for op_index in label_data:
                    op_index_new = self.op_sequence_offset_map.new_by_original[op_index]
                    op = self.op_sequence[op_index_new][0]
                    new_arg = op_index_to_arg(labeled_op_index_new)
                    if op in hasjrel:
                        new_arg = new_arg - (op_index_to_arg(op_index_new) + 2)
                    
                    self.set_arg(op_index_new, new_arg)
                
                labeled_op_index_new_after_value_change = self.op_sequence_offset_map.new_by_original[label]
                if labeled_op_index_new_after_value_change == labeled_op_index_new:
                    break
                else:
                    labeled_op_index_new = labeled_op_index_new_after_value_change

    def to_sequence_of_ints(self):
        for op, arg, _, _, _, _ in self.op_sequence:
            yield op
            if arg is None:
                yield 0
            else:
                yield arg

    def to_bytes(self) -> bytes:
        return bytes(self.to_sequence_of_ints())
        
    @staticmethod
    def from_bytecode_sequence(code: BytecodeSequence):
        labels, op_by_label = find_ops_with_labels(code)
        return OpSequence(list(unpack_opargs(code))), labels, op_by_label
    
    @staticmethod
    def from_entity(entity):
        code_bytes: BytecodeSequence = _get_code_object(entity).co_code
        return OpSequence.from_bytecode_sequence(code_bytes)
    
    @staticmethod
    def from_instructions(instructions: Sequence[Instruction]):
        op_sequence: List = list()
        for instruction in instructions:
            arg = instruction.arg
            if arg is None:
                arg = 0

            op_sequence.extend((instruction.opcode, arg))

        labels, op_by_label = find_ops_with_labels(op_sequence)
        result = OpSequence(list(unpack_opargs(op_sequence)))
        result.normalize_instructions_arg()
        return result, labels, op_by_label
    
    @staticmethod
    def from_instructions_fast(instructions: Sequence[Instruction]):
        return OpSequence(list(unpack_opargs(instructions_to_sequence_of_ints(normalize_instructions_arg(instructions)))))


def arg_to_bytes(arg: Union[int, bytes]) -> bytes:
    if isinstance(arg, int):
        arg = arg.to_bytes(4, byteorder='little')
        if len(arg) > 1:
            arg = arg.rstrip(b'\x00')
            
        if not arg:
            arg = b'\x00'
    
    return arg


def arg_to_int(arg: Union[int, bytes]) -> int:
    if isinstance(arg, bytes):
        arg = int.from_bytes(arg, byteorder='little')
    
    return arg


def get_raw_instructions(x, *, first_line=None):
    readable_instructions = get_instructions(x, first_line)
    for instruction in readable_instructions:
        arg = instruction.arg
        if arg is not None:
            if isinstance(arg, int):
                arg = arg.to_bytes(4, byteorder='little')
            
            arg = arg[0]

        yield Instruction(instruction.opname, instruction.opcode, arg, instruction.argval, instruction.argrepr, instruction.offset, instruction.starts_line, instruction.is_jump_target)


def normalize_instructions_arg(instructions: Sequence[Instruction]):
    for instruction in instructions:
        arg = instruction.arg
        if arg is not None:
            arg = arg_to_bytes(arg)
            extended_arg = arg[1:]
            arg = arg[0]
            for extended_arg_int in reversed(extended_arg):
                yield make_instruction('EXTENDED_ARG', extended_arg_int)

        yield Instruction(instruction.opname, instruction.opcode, arg, instruction.argval, instruction.argrepr, instruction.offset, instruction.starts_line, instruction.is_jump_target)


def instructions_to_sequence_of_ints(instructions: Sequence[Instruction]):
    for instruction in instructions:
        op = instruction.opcode
        yield op
        arg = instruction.arg
        yield 0 if arg is None else arg


def instructions_to_bytes(instructions: Sequence[Instruction]) -> bytes:
    return bytes(instructions_to_sequence_of_ints(instructions))


def make_instruction(name: str, arg: Union[None, int, bytes] = None) -> Instruction:
    name = name.upper()
    op = opmap[name]
    arg = arg_to_int(arg)
    return Instruction(name, op, arg, None, None, None, None, None)


mi = make_instruction


def fix_labels(op_sequence: OpSequence, op_by_label: Dict[int, List[Tuple[int, int, int, int, int, int]]]):
    op_sequence.fix_labels({arg_to_op_index(label): [op_info[2] for op_info in data] for label, data in op_by_label.items()})


def code_info(x):
    """Formatted details of methods, functions, or code."""
    return _format_code_info(_get_code_object(x))


def _pr(name, data):
    print(f'<<{name}>> type: {type(data)}; value: {data}')


def _format_code_info(co):
    _pr('Name', co.co_name)
    _pr('Filename', co.co_filename)
    _pr('Argument count', co.co_argcount)
    _pr('Positional-only arguments', co.co_posonlyargcount)
    _pr('Kw-only arguments', co.co_kwonlyargcount)
    _pr('Number of locals', co.co_nlocals)
    _pr('Stack size', co.co_stacksize)
    _pr('Flags', co.co_flags)
    _pr('Constants', co.co_consts)
    _pr('Names', co.co_names)
    _pr('Variable names', co.co_varnames)
    _pr('Free variables', co.co_freevars)
    _pr('Cell variables', co.co_cellvars)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import platform, sys
import codecs

"""
Determination Of the Correct Text Encoding
Obsolete. Use: `cengal/text_processing/encoding_detection` instead
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class BomInfo:
    def __init__(self, encodingName, strWithBOM):
        self.encodingName = encodingName
        self.strWithBOM = strWithBOM


class TextData:
    def __init__(self, bomInfo, text):
        self.bomInfo = bomInfo
        self.text = text


def parse_unicode_text(fData, isNeedToDecodeText=True):
    # utf8_BOM = bytes([0xef, 0xbb, 0xbf])
    # utf16LE_BOM = bytes([0xff, 0xfe])
    # utf16BE_BOM = bytes([0xfe, 0xff])
    utf8_BOM = codecs.BOM_UTF8
    utf16LE_BOM = codecs.BOM_UTF16_LE
    utf16BE_BOM = codecs.BOM_UTF16_BE
    spaceInUtf16LE = bytes([0x20, 0x00])
    spaceInUtf16BE = bytes([0x00, 0x20])
    isUTF = False
    isUTF16LE = False
    isUTF16BE = False
    isWithBOM = False

    #------------------
    #Preparing
    #------------------

    #preparing data: removing BOM if it exist
    #preparing triggers
    if fData.startswith(utf8_BOM):
        isUTF = True
        isWithBOM = True
        fData = fData[len(utf8_BOM):]
    elif fData.startswith(utf16LE_BOM):
        isUTF16LE = True
        isWithBOM = True
        fData = fData[len(utf16LE_BOM):]
    elif fData.startswith(utf16BE_BOM):
        isUTF16BE = True
        fData = fData[len(utf16BE_BOM):]
    elif (fData.count(spaceInUtf16BE) > 1) or (1 < fData.count(spaceInUtf16LE)):
        isWithBOM = False
        if (platform.system() == 'Windows') and (sys.byteorder == 'little'):
            isUTF16LE = True
        else:
            isUTF16BE = True

    #preparing file for editing: converting from UTF-16 if needed
    encodingName = 'utf-8'
    if isUTF:
        encodingName = 'utf-8'
    elif isUTF16LE:
        encodingName = 'utf-16le'
    elif isUTF16BE:
        encodingName = 'utf-16be'

    if isNeedToDecodeText:
        fData = str(fData, encodingName)

    #preparing BOM-trigger
    strWithBOM = b''
    if isWithBOM:
        if isUTF :
            strWithBOM = utf8_BOM
        elif isUTF16LE:
            strWithBOM = utf16LE_BOM
        elif isUTF16BE:
            strWithBOM = utf16BE_BOM

    bomInfo = BomInfo(encodingName, strWithBOM)
    textData = TextData(bomInfo, fData)
    return textData


def compile_unicode_text(textData):
    #preparing edited file for writing to disk: converting to UTF-16 if needed
    fData = bytes(textData.text, textData.bomInfo.encodingName)
    fData = textData.bomInfo.strWithBOM + fData
    return fData


def read_text_from_file(inputFile):
    textData = None
    with open(inputFile, "rb") as file:
        fData = file.read()
        textData = parse_unicode_text(fData)
    return textData


def write_text_to_file(outputFile, textData):
    #Writing edited file to disk
    with open(outputFile, "wb") as file:
        file.write(compile_unicode_text(textData))

# def remove_all_nonUtf8_symbols_from_utf8_text(binData, bytesFiller=b''):
#     data = parse_unicode_text(binData, False)
#     counter = 0
#     while counter < len(data.text):
#         if data.text[counter] > 0x7f:
#             # data.text[counter] = bytesFiller  # this is doesn't work
#             data.text = data.text[:counter] + bytesFiller + data.text[counter+1:]
#         else:
#             counter += 1
#     result = data.bomInfo.strWithBOM + data.text
#     return result


def remove_all_nonUtf8_symbols_from_utf8_text(binData):
    data = parse_unicode_text(binData, False)
    inputBinData = binData
    encodingName = 'utf-8'
    if data.bomInfo.encodingName.startswith('utf'):
        inputBinData = data.text
        encodingName = data.bomInfo.encodingName
    bufStr = inputBinData.decode(encodingName, 'replace')
    inputBinData = bufStr.encode(encodingName)

    result = data.bomInfo.strWithBOM + inputBinData
    return result


def get_BOM_from_text(binText):
    bomsList = (
        codecs.BOM
        , codecs.BOM_BE
        , codecs.BOM_LE
        , codecs.BOM32_BE
        , codecs.BOM32_LE
        , codecs.BOM64_BE
        , codecs.BOM64_LE
        , codecs.BOM_UTF8
        , codecs.BOM_UTF16
        , codecs.BOM_UTF16_BE
        , codecs.BOM_UTF16_LE
        , codecs.BOM_UTF32
        , codecs.BOM_UTF32_BE
        , codecs.BOM_UTF32_LE
    )

    result = (binText, None)

    for bom in bomsList:
        if binText.startswith(bom):
            result = (binText[len(bom):], bom)

    return result



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from cengal.text_processing.optional_formatter.versions.v_2 import OptionalFormatter, OptionalFormatterHandy, IT, AK
from enum import Enum

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class TimeAttributes(Enum):
    microseconds = 0
    milliseconds = 1
    seconds = 2
    minutes = 3
    hours = 4
    days = 5
    weeks = 6
    months = 7
    years = 8
    decades = 9
    centuries = 10
    millennia = 11


TA = TimeAttributes


class TestOptionalFormatter(unittest.TestCase):
    def setUp(self):
        self.formatter = OptionalFormatter((TA.years, TA.months, TA.days), {
            TA.years : IT('- ', '|(years)', '%{year}%', '|.', '; '),
            TA.months: IT('- ', '|(months)', '%{1}%', '|.', '; '),
            TA.days  : ('- ', '|(days)', '%{}%', '|.', '; ')
        })

    def test_all_parameters_are_present(self):
        result = self.formatter({
            TA.years: AK(1, "2", year=4, some=5),
            TA.months: AK(1, "2", 3, TA.months),
            TA.days : AK(1, "2", 3, TA.months, some=6)
        })
        expected_result = '|(years)%4%; - %2%; - %1%|.'
        self.assertEqual(result, expected_result, 'incorrect result')

    def test_some_parameters_are_present(self):
        result = self.formatter({
            TA.years: AK(1, "2", year=4, some=5),
            TA.days : AK(1, "2", 3, TA.months, some=6)
        })
        expected_result = '|(years)%4%; - %1%|.'
        self.assertEqual(result, expected_result, 'incorrect result')

    def test_none_of_parameters_are_present(self):
        result = self.formatter(dict())
        expected_result = ''
        self.assertEqual(result, expected_result, 'incorrect result')

    def test_none_of_parameters_are_present_2(self):
        result = self.formatter({})
        expected_result = ''
        self.assertEqual(result, expected_result, 'incorrect result')

    def test_blank_parameters_are_present_years(self):
        with self.assertRaises(KeyError) as cm:
            result = self.formatter({
                TA.years: AK()
            })

    def test_blank_parameters_are_present_months(self):
        with self.assertRaises(IndexError) as cm:
            result = self.formatter({
                TA.months: AK()
            })

    def test_blank_parameters_are_present_days(self):
        with self.assertRaises(IndexError) as cm:
            result = self.formatter({
                TA.days: AK()
            })

    def test_wrong_parameters_are_present_1(self):
        with self.subTest('1'):
            with self.assertRaises(KeyError) as cm:
                result = self.formatter({TA.years: AK(1, 2)})

    def test_wrong_parameters_are_present_2(self):
        with self.assertRaises(IndexError) as cm:
            result = self.formatter({TA.months: 1})

    def test_wrong_parameters_are_present_4(self):
        with self.assertRaises(TypeError) as cm:
            result = self.formatter(1)


if __name__ == '__main__':
    unittest.main()
